@@ -1,6 +1,6 @@
 /*
  *  R : A Computer Language for Statistical Data Analysis
- *  Copyright (C) 1997--2020  The R Core Team
+ *  Copyright (C) 1997--2023  The R Core Team
  *  Copyright (C) 1995, 1996  Robert Gentleman and Ross Ihaka
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -35,13 +35,18 @@
 #define R_imax2(x, y) ((x < y) ? y : x)
 #include <Print.h>
 
+#ifdef HAVE_SYS_TYPES_H
+# include <sys/types.h> // for size_t
+#endif
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
 
+#include <stdarg.h>
+
 #ifdef Win32
 void R_UTF8fixslash(char *s);
-static void R_wfixslash(wchar_t *s);
+void R_wfixslash(wchar_t *s);
 #endif
 
 #ifdef __cplusplus
@@ -53,10 +58,10 @@ extern "C" {
 #if defined FC_LEN_T
 # include <stddef.h>
 void F77_SYMBOL(rwarnc)(char *msg, int *nchar, FC_LEN_T msg_len);
-void NORET F77_SYMBOL(rexitc)(char *msg, int *nchar, FC_LEN_T msg_len);
+NORET void F77_SYMBOL(rexitc)(char *msg, int *nchar, FC_LEN_T msg_len);
 #else
 void F77_SYMBOL(rwarnc)(char *msg, int *nchar);
-void NORET F77_SYMBOL(rexitc)(char *msg, int *nchar);
+NORET void F77_SYMBOL(rexitc)(char *msg, int *nchar);
 #endif
 
 #ifdef __cplusplus
@@ -141,9 +146,9 @@ SEXP asChar(SEXP x)
 		if (LOGICAL(x)[0] == NA_LOGICAL)
 		    return NA_STRING;
 		if (LOGICAL(x)[0])
-		    sprintf(buf, "TRUE");
+		    snprintf(buf, MAXELTSIZE, "TRUE");
 		else
-		    sprintf(buf, "FALSE");
+		    snprintf(buf, MAXELTSIZE, "FALSE");
 		return mkChar(buf);
 	    case INTSXP:
 		if (INTEGER(x)[0] == NA_INTEGER)
@@ -324,7 +329,7 @@ const char *type2char(SEXPTYPE t) /* returns a char* */
 }
 
 #ifdef UNUSED
-SEXP NORET type2symbol(SEXPTYPE t)
+NORET SEXP type2symbol(SEXPTYPE t)
 {
     // if (t >= 0 && t < MAX_NUM_SEXPTYPE) { /* branch not really needed */
 	SEXP res = Type2Table[t].rsymName;
@@ -334,8 +339,8 @@ SEXP NORET type2symbol(SEXPTYPE t)
 }
 #endif
 
-attribute_hidden
-void NORET UNIMPLEMENTED_TYPEt(const char *s, SEXPTYPE t)
+attribute_hidden NORET
+void UNIMPLEMENTED_TYPEt(const char *s, SEXPTYPE t)
 {
     int i;
 
@@ -346,7 +351,7 @@ void NORET UNIMPLEMENTED_TYPEt(const char *s, SEXPTYPE t)
     error(_("unimplemented type (%d) in '%s'\n"), t, s);
 }
 
-void NORET UNIMPLEMENTED_TYPE(const char *s, SEXP x)
+NORET void UNIMPLEMENTED_TYPE(const char *s, SEXP x)
 {
     UNIMPLEMENTED_TYPEt(s, TYPEOF(x));
 }
@@ -417,7 +422,7 @@ Rboolean isBlankString(const char *s)
 	wchar_t wc; size_t used; mbstate_t mb_st;
 	mbs_init(&mb_st);
 	// This does not allow for surrogate pairs, but all blanks are in BMP
-	while( (used = Mbrtowc(&wc, s, MB_CUR_MAX, &mb_st)) ) {
+	while( (used = Mbrtowc(&wc, s, R_MB_CUR_MAX, &mb_st)) ) {
 	    if(!iswspace((wint_t) wc)) return FALSE;
 	    s += used;
 	}
@@ -454,7 +459,7 @@ Rboolean StringFalse(const char *name)
 }
 
 /* used in bind.c and options.c */
-SEXP attribute_hidden EnsureString(SEXP s)
+attribute_hidden SEXP EnsureString(SEXP s)
 {
     switch(TYPEOF(s)) {
     case SYMSXP:
@@ -493,7 +498,7 @@ void Rf_checkArityCall(SEXP op, SEXP args, SEXP call)
     }
 }
 
-void attribute_hidden Rf_check1arg(SEXP arg, SEXP call, const char *formal)
+attribute_hidden void Rf_check1arg(SEXP arg, SEXP call, const char *formal)
 {
     SEXP tag = TAG(arg);
     if (tag == R_NilValue) return;
@@ -525,7 +530,7 @@ SEXP R_listCompact(SEXP s, Rboolean keep_initial) {
     // skip initial NULL values
 	while (s != R_NilValue && CAR(s) == R_NilValue)
 	    s = CDR(s);
-    
+
     SEXP val = s;
     SEXP prev = s;
     while (s != R_NilValue) {
@@ -540,7 +545,7 @@ SEXP R_listCompact(SEXP s, Rboolean keep_initial) {
 
 
 /* This is a primitive (with no arguments) */
-SEXP attribute_hidden do_nargs(SEXP call, SEXP op, SEXP args, SEXP rho)
+attribute_hidden SEXP do_nargs(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     RCNTXT *cptr;
     int nargs = NA_INTEGER;
@@ -557,7 +562,7 @@ SEXP attribute_hidden do_nargs(SEXP call, SEXP op, SEXP args, SEXP rho)
 
 
 /* formerly used in subscript.c, in Utils.h */
-void attribute_hidden setIVector(int * vec, int len, int val)
+attribute_hidden void setIVector(int * vec, int len, int val)
 {
     for (int i = 0; i < len; i++) vec[i] = val;
 }
@@ -565,7 +570,7 @@ void attribute_hidden setIVector(int * vec, int len, int val)
 
 /* unused in R, in Utils.h, may have been used in Rcpp at some point,
       but not any more (as per Nov. 2018)  */
-void attribute_hidden setRVector(double * vec, int len, double val)
+attribute_hidden void setRVector(double * vec, int len, double val)
 {
     for (int i = 0; i < len; i++) vec[i] = val;
 }
@@ -648,7 +653,7 @@ SEXP R_body_no_src(SEXP x) {
    The return value is a list with 4 elements (xi, yi, x.alone, y.alone),
    which are index vectors for rows of x or y.
 */
-SEXP attribute_hidden do_merge(SEXP call, SEXP op, SEXP args, SEXP rho)
+attribute_hidden SEXP do_merge(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     SEXP xi, yi, ansx, ansy, ans;
     int nx = 0, ny = 0, i, j, k, nx_lone = 0, ny_lone = 0;
@@ -743,17 +748,22 @@ SEXP attribute_hidden do_merge(SEXP call, SEXP op, SEXP args, SEXP rho)
 # include <windows.h>
 #endif
 
+/* uses R_alloc */
 SEXP static intern_getwd(void)
 {
     SEXP rval = R_NilValue;
-    char buf[4*PATH_MAX+1];
 
 #ifdef Win32
     {
-	wchar_t wbuf[PATH_MAX+1];
-	int res = GetCurrentDirectoryW(PATH_MAX, wbuf);
-	if(res > 0) {
-	    wcstoutf8(buf, wbuf, sizeof(buf));
+	DWORD res = GetCurrentDirectoryW(0, NULL);
+	if (res > 0) {
+	    wchar_t *wbuf = (wchar_t*)R_alloc(res, sizeof(wchar_t));
+	    DWORD res1 = GetCurrentDirectoryW(res, wbuf);
+	    if (res1 <= 0 || res1 >= res)
+		return R_NilValue;
+	    size_t needed = wcstoutf8(NULL, wbuf, INT_MAX);
+	    char *buf = R_alloc(needed + 1, 1);
+	    wcstoutf8(buf, wbuf, needed + 1);
 	    R_UTF8fixslash(buf);
 	    PROTECT(rval = allocVector(STRSXP, 1));
 	    SET_STRING_ELT(rval, 0, mkCharCE(buf, CE_UTF8));
@@ -761,13 +771,14 @@ SEXP static intern_getwd(void)
 	}
     }
 #else
-    char *res = getcwd(buf, PATH_MAX); /* can return NULL */
+    char buf[4*R_PATH_MAX+1];
+    char *res = getcwd(buf, R_PATH_MAX); /* can return NULL */
     if(res) rval = mkString(buf);
 #endif
     return(rval);
 }
 
-SEXP attribute_hidden do_getwd(SEXP call, SEXP op, SEXP args, SEXP rho)
+attribute_hidden SEXP do_getwd(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     checkArity(op, args);
 
@@ -779,7 +790,7 @@ SEXP attribute_hidden do_getwd(SEXP call, SEXP op, SEXP args, SEXP rho)
 # include <direct.h> /* for chdir, via io.h */
 #endif
 
-SEXP attribute_hidden do_setwd(SEXP call, SEXP op, SEXP args, SEXP rho)
+attribute_hidden SEXP do_setwd(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     SEXP s = R_NilValue, wd = R_NilValue;	/* -Wall */
 
@@ -813,11 +824,11 @@ SEXP attribute_hidden do_setwd(SEXP call, SEXP op, SEXP args, SEXP rho)
 /* remove portion of path before file separator if one exists */
 
 #ifdef Win32
-SEXP attribute_hidden do_basename(SEXP call, SEXP op, SEXP args, SEXP rho)
+attribute_hidden SEXP do_basename(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     SEXP ans, s = R_NilValue;	/* -Wall */
-    char sp[4*PATH_MAX+1];
-    wchar_t  buf[PATH_MAX], *p;
+    char *sp;
+    wchar_t  *buf, *p;
     const wchar_t *pp;
     int i, n;
 
@@ -830,16 +841,20 @@ SEXP attribute_hidden do_basename(SEXP call, SEXP op, SEXP args, SEXP rho)
 	    SET_STRING_ELT(ans, i, NA_STRING);
 	else {
 	    pp = filenameToWchar(STRING_ELT(s, i), TRUE);
-	    if (wcslen(pp) > PATH_MAX - 1) error(_("path too long"));
+	    buf = (wchar_t *)R_alloc(wcslen(pp) + 1, sizeof(wchar_t));
 	    wcscpy(buf, pp);
 	    R_wfixslash(buf);
 	    /* remove trailing file separator(s) */
 	    if (*buf) {
 		p = buf + wcslen(buf) - 1;
+		/* turns D:/ to D: */
+		/* FIXME: basename of D:/ is D:, is that a good behavior? */
 		while (p >= buf && *p == L'/') *(p--) = L'\0';
 	    }
 	    if ((p = wcsrchr(buf, L'/'))) p++; else p = buf;
-	    wcstoutf8(sp, p, sizeof(sp));
+	    size_t needed = wcstoutf8(NULL, p, INT_MAX);
+	    sp = R_alloc(needed + 1, 1);
+	    wcstoutf8(sp, p, needed + 1);
 	    SET_STRING_ELT(ans, i, mkCharCE(sp, CE_UTF8));
 	}
     }
@@ -847,10 +862,10 @@ SEXP attribute_hidden do_basename(SEXP call, SEXP op, SEXP args, SEXP rho)
     return(ans);
 }
 #else
-SEXP attribute_hidden do_basename(SEXP call, SEXP op, SEXP args, SEXP rho)
+attribute_hidden SEXP do_basename(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     SEXP ans, s = R_NilValue;	/* -Wall */
-    char  buf[PATH_MAX], *p, fsp = FILESEP[0];
+    char  buf[R_PATH_MAX], *p, fsp = FILESEP[0];
     const char *pp;
     int i, n;
 
@@ -863,7 +878,7 @@ SEXP attribute_hidden do_basename(SEXP call, SEXP op, SEXP args, SEXP rho)
 	    SET_STRING_ELT(ans, i, NA_STRING);
 	else {
 	    pp = R_ExpandFileName(translateCharFP(STRING_ELT(s, i)));
-	    if (strlen(pp) > PATH_MAX - 1)
+	    if (strlen(pp) > R_PATH_MAX - 1)
 		error(_("path too long"));
 	    strcpy (buf, pp);
 	    if (*buf) {
@@ -887,12 +902,12 @@ SEXP attribute_hidden do_basename(SEXP call, SEXP op, SEXP args, SEXP rho)
    */
 
 #ifdef Win32
-SEXP attribute_hidden do_dirname(SEXP call, SEXP op, SEXP args, SEXP rho)
+attribute_hidden SEXP do_dirname(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     SEXP ans, s = R_NilValue;	/* -Wall */
-    wchar_t buf[PATH_MAX], *p;
+    wchar_t *buf, *p;
     const wchar_t *pp;
-    char sp[4*PATH_MAX+1];
+    char *sp;
     int i, n;
 
     checkArity(op, args);
@@ -903,18 +918,18 @@ SEXP attribute_hidden do_dirname(SEXP call, SEXP op, SEXP args, SEXP rho)
 	if (STRING_ELT(s, i) == NA_STRING)
 	    SET_STRING_ELT(ans, i, NA_STRING);
 	else {
-	    memset(sp, 0, 4*PATH_MAX);
+	    sp = "";
 	    pp = filenameToWchar(STRING_ELT(s, i), TRUE);
-	    if (wcslen(pp) > PATH_MAX - 1)
-		error(_("path too long"));
 	    if (wcslen(pp)) {
+		buf = (wchar_t*)R_alloc(wcslen(pp) + 1, sizeof(wchar_t));
 		wcscpy (buf, pp);
 		R_wfixslash(buf);
-		/* remove trailing file separator(s) */
+		/* remove trailing file separator(s), preserve D:/, / */
 		p = buf + wcslen(buf) - 1;
 		while (p > buf && *p == L'/'
 		       && (p > buf+2 || *(p-1) != L':')) *p-- = L'\0';
 		p = wcsrchr(buf, L'/');
+		/* FIXME: dirname of D: is ., is this a good behavior? */
 		if(p == NULL) wcscpy(buf, L".");
 		else {
 		    while(p > buf && *p == L'/'
@@ -922,7 +937,9 @@ SEXP attribute_hidden do_dirname(SEXP call, SEXP op, SEXP args, SEXP rho)
 			  && (p > buf+2 || *(p-1) != L':')) --p;
 		    p[1] = L'\0';
 		}
-		wcstoutf8(sp, buf, sizeof(sp));
+		size_t needed = wcstoutf8(NULL, buf, INT_MAX);
+		sp = R_alloc(needed + 1, 1);
+		wcstoutf8(sp, buf, needed + 1);
 	    }
 	    SET_STRING_ELT(ans, i, mkCharCE(sp, CE_UTF8));
 	}
@@ -931,10 +948,10 @@ SEXP attribute_hidden do_dirname(SEXP call, SEXP op, SEXP args, SEXP rho)
     return(ans);
 }
 #else
-SEXP attribute_hidden do_dirname(SEXP call, SEXP op, SEXP args, SEXP rho)
+attribute_hidden SEXP do_dirname(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     SEXP ans, s = R_NilValue;	/* -Wall */
-    char buf[PATH_MAX], *p, fsp = FILESEP[0];
+    char buf[R_PATH_MAX], *p, fsp = FILESEP[0];
     const char *pp;
     int i, n;
 
@@ -947,7 +964,7 @@ SEXP attribute_hidden do_dirname(SEXP call, SEXP op, SEXP args, SEXP rho)
 	    SET_STRING_ELT(ans, i, NA_STRING);
 	else {
 	    pp = R_ExpandFileName(translateCharFP(STRING_ELT(s, i)));
-	    if (strlen(pp) > PATH_MAX - 1)
+	    if (strlen(pp) > R_PATH_MAX - 1)
 		error(_("path too long"));
 	    size_t ll = strlen(pp);
 	    if (ll) { // svMisc calls this with ""
@@ -976,12 +993,12 @@ SEXP attribute_hidden do_dirname(SEXP call, SEXP op, SEXP args, SEXP rho)
 extern char *realpath(const char *path, char *resolved_path);
 #endif
 
-SEXP attribute_hidden do_normalizepath(SEXP call, SEXP op, SEXP args, SEXP rho)
+attribute_hidden SEXP do_normalizepath(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     SEXP ans, paths = CAR(args), elp;
     int i, n = LENGTH(paths);
     const char *path;
-    char abspath[PATH_MAX+1];
+    char abspath[R_PATH_MAX+1];
 
     checkArity(op, args);
     if (!isString(paths))
@@ -1034,12 +1051,12 @@ SEXP attribute_hidden do_normalizepath(SEXP call, SEXP op, SEXP args, SEXP rho)
 	    continue;
 	}
 	path = translateCharFP(elp);
-	OK = strlen(path) <= PATH_MAX;
+	OK = strlen(path) <= R_PATH_MAX;
 	if (OK) {
-	    if (path[0] == '/') strncpy(abspath, path, PATH_MAX);
+	    if (path[0] == '/') strncpy(abspath, path, R_PATH_MAX);
 	    else {
-		OK = getcwd(abspath, PATH_MAX) != NULL;
-		OK = OK && (strlen(path) + strlen(abspath) + 1 <= PATH_MAX);
+		OK = getcwd(abspath, R_PATH_MAX) != NULL;
+		OK = OK && (strlen(path) + strlen(abspath) + 1 <= R_PATH_MAX);
 		if (OK) {strcat(abspath, "/"); strcat(abspath, path);}
 	    }
 	}
@@ -1063,7 +1080,8 @@ SEXP attribute_hidden do_normalizepath(SEXP call, SEXP op, SEXP args, SEXP rho)
 #ifdef USE_INTERNAL_MKTIME
 const char *getTZinfo(void)
 {
-    static char def_tz[PATH_MAX+1] = "";
+    /* FIXME: use filesystem-agnostic limit?*/
+    static char def_tz[R_PATH_MAX+1] = "";
     if (def_tz[0]) return def_tz;
 
     // call Sys.timezone()
@@ -1073,6 +1091,8 @@ const char *getTZinfo(void)
     if(TYPEOF(ans) == STRSXP && LENGTH(ans) == 1) {
 	SEXP el = STRING_ELT(ans, 0);
 	if (el != NA_STRING) {
+	    if (strlen(CHAR(el)) + 1 > sizeof(def_tz))
+		error("time zone specification is too long");
 	    strcpy(def_tz, CHAR(el));
 	    // printf("tz is %s\n", CHAR(el));
 	    UNPROTECT(3);
@@ -1097,7 +1117,7 @@ static void encode_cleanup(void *data)
 #endif
 
 /* encodeString(x, w, quote, justify) */
-SEXP attribute_hidden do_encodeString(SEXP call, SEXP op, SEXP args, SEXP rho)
+attribute_hidden SEXP do_encodeString(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     SEXP ans, x, s;
     R_xlen_t i, len;
@@ -1178,7 +1198,7 @@ SEXP attribute_hidden do_encodeString(SEXP call, SEXP op, SEXP args, SEXP rho)
     return ans;
 }
 
-SEXP attribute_hidden do_encoding(SEXP call, SEXP op, SEXP args, SEXP rho)
+attribute_hidden SEXP do_encoding(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     SEXP ans, x;
     R_xlen_t i, n;
@@ -1203,7 +1223,7 @@ SEXP attribute_hidden do_encoding(SEXP call, SEXP op, SEXP args, SEXP rho)
 #define IS_NATIVE(tmp) \
     (! IS_LATIN1(tmp) && ! IS_UTF8(tmp) && ! IS_BYTES(tmp))
 
-SEXP attribute_hidden do_setencoding(SEXP call, SEXP op, SEXP args, SEXP rho)
+attribute_hidden SEXP do_setencoding(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     SEXP x, enc, tmp;
     int m;
@@ -1239,7 +1259,9 @@ SEXP attribute_hidden do_setencoding(SEXP call, SEXP op, SEXP args, SEXP rho)
     return x;
 }
 
-SEXP attribute_hidden markKnown(const char *s, SEXP ref)
+/* `*s` should point to a string derived from `ref` after `ref` has been
+   translated to native encoding.  See `?Encoding` */
+attribute_hidden SEXP markKnown(const char *s, SEXP ref)
 {
     int ienc = 0;
     if(ENC_KNOWN(ref)) {
@@ -1304,9 +1326,11 @@ utf8toucs(wchar_t *wc, const char *s)
     if (byte == 0) {
 	*w = (wchar_t) 0;
 	return 0;
-    } else if (byte < 0xC0) {
+    } else if (byte < 0x80) {
 	*w = (wchar_t) byte;
 	return 1;
+    } else if (byte < 0xC0) {
+	return (size_t)-1;
     } else if (byte < 0xE0) {
 	if(strlen(s) < 2) return (size_t)-2;
 	if ((s[1] & 0xC0) == 0x80) {
@@ -1406,8 +1430,11 @@ utf8towcs4(R_wchar_t *wc, const char *s, size_t n)
 
     if(wc)
 	for(p = wc, t = s; ; p++, t += m) {
-	    // FIXME this gives a warning on Windows.
-	    m  = (ssize_t) utf8toucs(p, t);
+	    wchar_t local = L'\0';
+	    m  = (ssize_t) utf8toucs(&local, t);
+	    /* Needed to write all of R_wchar_t even on Windows where it is bigger
+	       than wchar_t */
+	    *p = (R_wchar_t) local;
 	    if (m < 0) error(_("invalid input '%s' in 'utf8towcs32'"), s);
 	    if (m == 0) break;
 	    if (IS_HIGH_SURROGATE(*p)) *p = utf8toucs32(*p, s);
@@ -1463,7 +1490,7 @@ static size_t Rwcrtomb32(char *s, R_wchar_t cvalue, size_t n)
 
    The return value is the number of chars including the terminating
    null.  If the buffer is not big enough, the result is truncated but
-   still null-terminated 
+   still null-terminated
 */
 attribute_hidden // but used in windlgs
 size_t wcstoutf8(char *s, const wchar_t *wc, size_t n)
@@ -1520,7 +1547,8 @@ size_t Mbrtowc(wchar_t *wc, const char *s, size_t n, mbstate_t *ps)
 	if (!R_Is_Running) return (size_t)-1;
 	/* let's try to print out a readable version */
 	R_CheckStack2(4*strlen(s) + 10);
-	char err[4*strlen(s) + 1], *q;
+	size_t sz = 4*strlen(s) + 1;
+	char err[sz], *q;
 	const char *p;
 	for(p = s, q = err; *p; ) {
 	    /* don't do the first to keep ps state straight */
@@ -1529,11 +1557,11 @@ size_t Mbrtowc(wchar_t *wc, const char *s, size_t n, mbstate_t *ps)
 	    else if((int) used > 0) {
 		memcpy(q, p, used);
 		p += used;
-		q += used;
+		q += used; sz -= used;
 		n -= used;
 	    } else {
-		sprintf(q, "<%02x>", (unsigned char) *p++);
-		q += 4;
+		snprintf(q, sz, "<%02x>", (unsigned char) *p++);
+		q += 4; sz -= 4;
 		n--;
 	    }
 	}
@@ -1544,7 +1572,7 @@ size_t Mbrtowc(wchar_t *wc, const char *s, size_t n, mbstate_t *ps)
 }
 
 /* Truncate a string in place (in native encoding) so that it only contains
-   valid multi-byte characters. Has no effect in non-mbcs locales. 
+   valid multi-byte characters. Has no effect in non-mbcs locales.
 
    This function may be invoked by the error handler via
    REvprintf->Rvsnprintf_mbcs.  Do not change it unless you are SURE that
@@ -1600,7 +1628,7 @@ Rboolean utf8Valid(const char *str)
     return valid_utf8(str, strlen(str)) == 0;
 }
 
-SEXP attribute_hidden do_validUTF8(SEXP call, SEXP op, SEXP args, SEXP rho)
+attribute_hidden SEXP do_validUTF8(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     checkArity(op, args);
     SEXP x = CAR(args);
@@ -1614,7 +1642,7 @@ SEXP attribute_hidden do_validUTF8(SEXP call, SEXP op, SEXP args, SEXP rho)
     return ans;
 }
 
-SEXP attribute_hidden do_validEnc(SEXP call, SEXP op, SEXP args, SEXP rho)
+attribute_hidden SEXP do_validEnc(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     checkArity(op, args);
     SEXP x = CAR(args);
@@ -1643,7 +1671,7 @@ char *Rf_strchr(const char *s, int c)
 
     if(!mbcslocale || utf8locale) return strchr(s, c);
     mbs_init(&mb_st);
-    while( (used = Mbrtowc(NULL, p, MB_CUR_MAX, &mb_st)) ) {
+    while( (used = Mbrtowc(NULL, p, R_MB_CUR_MAX, &mb_st)) ) {
 	if(*p == c) return p;
 	p += used;
     }
@@ -1658,7 +1686,7 @@ char *Rf_strrchr(const char *s, int c)
 
     if(!mbcslocale || utf8locale) return strrchr(s, c);
     mbs_init(&mb_st);
-    while( (used = Mbrtowc(NULL, p, MB_CUR_MAX, &mb_st)) ) {
+    while( (used = Mbrtowc(NULL, p, R_MB_CUR_MAX, &mb_st)) ) {
 	if(*p == c) plast = p;
 	p += used;
     }
@@ -1673,7 +1701,7 @@ void R_fixslash(char *s)
     if(mbcslocale) {
 	mbstate_t mb_st; int used;
 	mbs_init(&mb_st);
-	while((used = Mbrtowc(NULL, p, MB_CUR_MAX, &mb_st))) {
+	while((used = Mbrtowc(NULL, p, R_MB_CUR_MAX, &mb_st))) {
 	    if(*p == '\\') *p = '/';
 	    p += used;
 	}
@@ -1687,12 +1715,12 @@ void R_UTF8fixslash(char *s)
 {
     char *p = s;
 
-	for (; *p; p++) if (*p == '\\') *p = '/';
-	/* preserve network shares */
-	if(s[0] == '/' && s[1] == '/') s[0] = s[1] = '\\';
+    for (; *p; p++) if (*p == '\\') *p = '/';
+    /* preserve network shares */
+    if(s[0] == '/' && s[1] == '/') s[0] = s[1] = '\\';
 }
 
-static void R_wfixslash(wchar_t *s)
+void R_wfixslash(wchar_t *s)
 {
     wchar_t *p = s;
 
@@ -1701,7 +1729,6 @@ static void R_wfixslash(wchar_t *s)
     if(s[0] == L'/' && s[1] == L'/') s[0] = s[1] = L'\\';
 }
 
-
 void R_fixbackslash(char *s)
 {
     char *p = s;
@@ -1709,19 +1736,27 @@ void R_fixbackslash(char *s)
     if(mbcslocale) {
 	mbstate_t mb_st; int used;
 	mbs_init(&mb_st);
-	while((used = Mbrtowc(NULL, p, MB_CUR_MAX, &mb_st))) {
+	while((used = Mbrtowc(NULL, p, R_MB_CUR_MAX, &mb_st))) {
 	    if(*p == '/') *p = '\\';
 	    p += used;
 	}
     } else
 	for (; *p; p++) if (*p == '/') *p = '\\';
 }
+
+void R_wfixbackslash(wchar_t *s)
+{
+    wchar_t *p = s;
+
+    for (; *p; p++) if (*p == L'/') *p = L'\\';
+}
+
 #endif
 
 #if defined FC_LEN_T
-void NORET F77_SYMBOL(rexitc)(char *msg, int *nchar, FC_LEN_T msg_len)
+NORET void F77_SYMBOL(rexitc)(char *msg, int *nchar, FC_LEN_T msg_len)
 #else
-void NORET F77_SYMBOL(rexitc)(char *msg, int *nchar)
+NORET void F77_SYMBOL(rexitc)(char *msg, int *nchar)
 #endif
 {
     int nc = *nchar;
@@ -2155,7 +2190,7 @@ double R_atof(const char *str)
 
 /* enc2native and enc2utf8, but they are the same in a UTF-8 locale */
 /* primitive */
-SEXP attribute_hidden do_enc2(SEXP call, SEXP op, SEXP args, SEXP env)
+attribute_hidden SEXP do_enc2(SEXP call, SEXP op, SEXP args, SEXP env)
 {
     SEXP ans, el;
     R_xlen_t i;
@@ -2287,7 +2322,7 @@ static UCollator *collator = NULL;
 static int collationLocaleSet = 0;
 
 /* called from platform.c */
-void attribute_hidden resetICUcollator(Rboolean disable)
+attribute_hidden void resetICUcollator(Rboolean disable)
 {
     if (collator) ucol_close(collator);
     collator = NULL;
@@ -2305,8 +2340,8 @@ static const struct {
     { "strength", 999 },
     { "primary ", UCOL_PRIMARY },
     { "secondary ", UCOL_SECONDARY },
-    { "teritary ", UCOL_TERTIARY },
-    { "guaternary ", UCOL_QUATERNARY },
+    { "tertiary ", UCOL_TERTIARY },
+    { "quaternary ", UCOL_QUATERNARY },
     { "identical ", UCOL_IDENTICAL },
     { "french_collation", UCOL_FRENCH_COLLATION },
     { "on", UCOL_ON },
@@ -2329,19 +2364,19 @@ static const char *getLocale(void)
     const char *p = getenv("R_ICU_LOCALE");
     if (p && p[0]) return p;
 
-    // This call is >= Vista/Server 2008
+    // FIXME: ideally, we would use a locale name corresponding to the current
+    //        C runtime locale, so as reported by setlocale(); since 4.2 this
+    //        means a UCRT locale name. As this is only approximated, we don't
+    //        use ICU by default on Windows yet for collation, even though
+    //        already having UTF-8 as the native encoding.
+
     // ICU should accept almost all of these, e.g. en-US and uz-Latn-UZ
-    PGSDLN pGSDLN = (PGSDLN)
-	GetProcAddress(GetModuleHandle(TEXT("kernel32.dll")),
-		       "GetSystemDefaultLocaleName");
-    if(pGSDLN) {
-	WCHAR wcBuffer[BUFFER_SIZE];
-	pGSDLN(wcBuffer, BUFFER_SIZE);
-	static char locale[BUFFER_SIZE];
-	WideCharToMultiByte(CP_ACP, 0, wcBuffer, -1,
-			    locale, BUFFER_SIZE, NULL, NULL);
-	return locale;
-    } else return "root";
+    WCHAR wcBuffer[BUFFER_SIZE];
+    GetSystemDefaultLocaleName(wcBuffer, BUFFER_SIZE);
+    static char locale[BUFFER_SIZE];
+    WideCharToMultiByte(CP_ACP, 0, wcBuffer, -1,
+			locale, BUFFER_SIZE, NULL, NULL);
+    return locale;
 }
 #else
 static const char *getLocale(void)
@@ -2351,7 +2386,7 @@ static const char *getLocale(void)
 }
 #endif
 
-SEXP attribute_hidden do_ICUset(SEXP call, SEXP op, SEXP args, SEXP rho)
+attribute_hidden SEXP do_ICUset(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     SEXP x;
     UErrorCode  status = U_ZERO_ERROR;
@@ -2373,7 +2408,16 @@ SEXP attribute_hidden do_ICUset(SEXP call, SEXP op, SEXP args, SEXP rho)
 	    if(streql(s, "ASCII")) {
 		collationLocaleSet = 2;
 	    } else {
-		if(strcmp(s, "none")) {
+		int usable_icu = 1;
+#ifdef Win32
+		/* ICU 72 requires this function (and other from Windows 7) */
+		if (!GetProcAddress(GetModuleHandle(TEXT("kernel32")),
+		                    "ResolveLocaleName")) {
+		    usable_icu = 0;
+		    warning("cannot use ICU on this system");
+		}
+#endif
+		if(usable_icu && strcmp(s, "none")) {
 		    if(streql(s, "default"))
 			uloc_setDefault(getLocale(), &status);
 		    else uloc_setDefault(s, &status);
@@ -2412,7 +2456,7 @@ SEXP attribute_hidden do_ICUset(SEXP call, SEXP op, SEXP args, SEXP rho)
     return R_NilValue;
 }
 
-SEXP attribute_hidden do_ICUget(SEXP call, SEXP op, SEXP args, SEXP rho)
+attribute_hidden SEXP do_ICUget(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     const char *ans = "unknown", *res;
     checkArity(op, args);
@@ -2461,7 +2505,16 @@ int Scollate(SEXP a, SEXP b)
 	/* FIXME: as ICU does not support C as locale, could we use the Unix
 	   behavior on all systems? */
 	const char *p = getenv("R_ICU_LOCALE");
-	if(p && p[0] && (!useC || !strcmp(p, "C"))) {
+	int use_icu = p && p[0] && (!useC || !strcmp(p, "C"));
+
+	/* ICU 72 requires this function (and other from Windows 7) */
+	if (use_icu &&
+	    !GetProcAddress(GetModuleHandle(TEXT("kernel32")),
+			    "ResolveLocaleName")) {
+	    use_icu = 0;
+	    warning("cannot use ICU on this system");
+	}
+	if(use_icu) {
 #endif
 	    UErrorCode status = U_ZERO_ERROR;
 	    uloc_setDefault(getLocale(), &status);
@@ -2494,19 +2547,19 @@ int Scollate(SEXP a, SEXP b)
 
 #else /* not USE_ICU */
 
-SEXP attribute_hidden do_ICUset(SEXP call, SEXP op, SEXP args, SEXP rho)
+attribute_hidden SEXP do_ICUset(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     warning(_("ICU is not supported on this build"));
     return R_NilValue;
 }
 
-SEXP attribute_hidden do_ICUget(SEXP call, SEXP op, SEXP args, SEXP rho)
+attribute_hidden SEXP do_ICUget(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     checkArity(op, args);
     return mkString("ICU not in use");
 }
 
-void attribute_hidden resetICUcollator(Rboolean disable) {}
+attribute_hidden void resetICUcollator(Rboolean disable) {}
 
 # ifdef Win32
 
@@ -2539,7 +2592,7 @@ int Scollate(SEXP a, SEXP b)
 
 #include <lzma.h>
 
-SEXP attribute_hidden do_crc64(SEXP call, SEXP op, SEXP args, SEXP rho)
+attribute_hidden SEXP do_crc64(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     checkArity(op, args);
     SEXP in = CAR(args);
@@ -2587,7 +2640,7 @@ bincode(double *x, R_xlen_t n, double *breaks, int nb,
 }
 
 /* 'breaks' cannot be a long vector as the return codes are integer. */
-SEXP attribute_hidden do_bincode(SEXP call, SEXP op, SEXP args, SEXP rho)
+attribute_hidden SEXP do_bincode(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     checkArity(op, args);
     SEXP x, breaks, right, lowest;
@@ -2613,7 +2666,7 @@ SEXP attribute_hidden do_bincode(SEXP call, SEXP op, SEXP args, SEXP rho)
     return codes;
 }
 
-SEXP attribute_hidden do_tabulate(SEXP call, SEXP op, SEXP args, SEXP rho)
+attribute_hidden SEXP do_tabulate(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     checkArity(op, args);
     SEXP in = CAR(args), nbin = CADR(args);
@@ -2648,7 +2701,7 @@ SEXP attribute_hidden do_tabulate(SEXP call, SEXP op, SEXP args, SEXP rho)
  *                         xt  x    right             inside       leftOp
  * x can be a long vector but xt cannot since the result is integer
 */
-SEXP attribute_hidden do_findinterval(SEXP call, SEXP op, SEXP args, SEXP rho)
+attribute_hidden SEXP do_findinterval(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     checkArity(op, args);
     SEXP xt, x, right, inside, leftOp;
@@ -2690,10 +2743,12 @@ SEXP attribute_hidden do_findinterval(SEXP call, SEXP op, SEXP args, SEXP rho)
 # undef ERROR
 #endif
 #include <R_ext/Applic.h>
-SEXP attribute_hidden do_pretty(SEXP call, SEXP op, SEXP args, SEXP rho)
+/* .Internal(pretty(min(x), max(x), n, min.n, shrink.sml,
+ *                  c(high.u.bias, u5.bias), eps.correct))
+ */
+attribute_hidden SEXP do_pretty(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     checkArity(op, args);
-    SEXP ans, nm, hi;
     double l = asReal(CAR(args)); args = CDR(args);
     if (!R_FINITE(l)) error(_("invalid '%s' argument"), "l");
     double u = asReal(CAR(args)); args = CDR(args);
@@ -2706,26 +2761,42 @@ SEXP attribute_hidden do_pretty(SEXP call, SEXP op, SEXP args, SEXP rho)
     double shrink = asReal(CAR(args)); args = CDR(args);
     if (!R_FINITE(shrink) || shrink <= 0.)
 	error(_("invalid '%s' argument"), "shrink.sml");
-    PROTECT(hi = coerceVector(CAR(args), REALSXP)); args = CDR(args);
-    double z;
-    if (!R_FINITE(z = REAL(hi)[0]) || z < 0.)
+    SEXP hi = PROTECT(coerceVector(CAR(args), REALSXP)); args = CDR(args); // (h, h5, f.min)
+    double *z = REAL(hi);
+    if (!R_FINITE(z[0]) || z[0] < 0.)
 	error(_("invalid '%s' argument"), "high.u.bias");
-    if (!R_FINITE(z = REAL(hi)[1]) || z < 0.)
+    if (!R_FINITE(z[1]) || z[1] < 0.)
 	error(_("invalid '%s' argument"), "u5.bias");
-    int eps = asInteger(CAR(args)); /* eps.correct */
+    if (!R_FINITE(z[2]) || z[2] <= 0.)
+	error(_("invalid '%s' argument"), "f.min");
+    int eps = asInteger(CAR(args)); args = CDR(args); /* eps.correct */
     if (eps == NA_INTEGER || eps < 0 || eps > 2)
 	error(_("'eps.correct' must be 0, 1, or 2"));
-    R_pretty(&l, &u, &n, min_n, shrink, REAL(hi), eps, 1);
-    //------ (returns 'unit' which we do not need)
-    PROTECT(ans = allocVector(VECSXP, 3));
+    int return_bounds = asLogical(CAR(args)); args = CDR(args); /* bounds */
+    if (return_bounds == NA_LOGICAL)
+	error(_("'bounds' must be TRUE or FALSE"));
+    double unit;
+    if(return_bounds)
+	       R_pretty(&l, &u, &n, min_n, shrink, REAL(hi), eps, 1);
+    else // unit  and (ns,nu)
+	unit = R_pretty(&l, &u, &n, min_n, shrink, REAL(hi), eps, 0);
+    int l_ans = return_bounds ? 3 : 4;
+    SEXP ans = PROTECT(allocVector(VECSXP, l_ans)),
+	nm = allocVector(STRSXP, l_ans);
+    setAttrib(ans, R_NamesSymbol, nm);
     SET_VECTOR_ELT(ans, 0, ScalarReal(l));
     SET_VECTOR_ELT(ans, 1, ScalarReal(u));
     SET_VECTOR_ELT(ans, 2, ScalarInteger(n));
-    nm = allocVector(STRSXP, 3);
-    setAttrib(ans, R_NamesSymbol, nm);
-    SET_STRING_ELT(nm, 0, mkChar("l"));
-    SET_STRING_ELT(nm, 1, mkChar("u"));
     SET_STRING_ELT(nm, 2, mkChar("n"));
+    if(return_bounds) {
+	SET_STRING_ELT(nm, 0, mkChar("l"));
+	SET_STRING_ELT(nm, 1, mkChar("u"));
+    } else {
+	SET_STRING_ELT(nm, 0, mkChar("ns"));
+	SET_STRING_ELT(nm, 1, mkChar("nu"));
+	SET_STRING_ELT(nm, 3, mkChar("unit"));
+	SET_VECTOR_ELT(ans,3, ScalarReal(unit));
+    }
     UNPROTECT(2);
     return ans;
 }
@@ -2739,7 +2810,7 @@ static void
 str_signif_sexp(SEXP x, const char *type, int width, int digits,
 	   const char *format, const char *flag, char **result);
 
-SEXP attribute_hidden do_formatC(SEXP call, SEXP op, SEXP args, SEXP rho)
+attribute_hidden SEXP do_formatC(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     checkArity(op, args);
     SEXP x = CAR(args); args = CDR(args);
@@ -2993,3 +3064,107 @@ char *Rstrdup(const char *s)
     memcpy (cpy, s, nb);
     return (char *) cpy;
 }
+
+static int compareNumericVersion(SEXP x, SEXP y)
+{
+    int i, nx, ny, nc, *ix, *iy;
+    if(!isInteger(x))
+	error(_("invalid '%s' argument"), "x");
+    if(!isInteger(y))
+	error(_("invalid '%s' argument"), "y");
+    nx = LENGTH(x);
+    ny = LENGTH(y);
+    nc =  nx > ny ? ny : nx;
+    if(nc == 0)
+	return NA_INTEGER;
+    ix = INTEGER(x);
+    iy = INTEGER(y);
+    for(i = 0; i < nc; i++) {
+	if(ix[i] > iy[i])
+	    return 1;
+	if(ix[i] < iy[i])
+	    return -1;
+    }
+    if(nc < nx) {
+	for(i = nc; i < nx; i++) {
+	    if(ix[i] > 0)
+		return 1;
+	}
+    } else if(nc < ny) {
+	for(i = nc; i < ny; i++) {
+	    if(iy[i] > 0)
+		return -1;
+	}
+    }
+    return 0;
+}
+
+attribute_hidden
+SEXP do_compareNumericVersion(SEXP call, SEXP op, SEXP args, SEXP env)
+{
+    SEXP ans, x, y;
+    int i, nx, ny, na;
+
+    checkArity(op, args);
+
+    x = CAR(args); args = CDR(args);
+    y = CAR(args);
+
+    if(!isVector(x))
+	error(_("invalid '%s' argument"), "x");
+    if(!isVector(y))
+	error(_("invalid '%s' argument"), "y");
+    nx = LENGTH(x);
+    ny = LENGTH(y);
+    if(nx > 0 && ny > 0)
+	na = nx > ny ? nx : ny;
+    else
+	na = 0;
+    PROTECT(ans = allocVector(INTSXP, na));
+    for(i = 0; i < na; i++) {
+	INTEGER(ans)[i] = 
+	    compareNumericVersion(VECTOR_ELT(x, i % nx),
+				  VECTOR_ELT(y, i % ny));
+    }
+    UNPROTECT(1);
+    return ans;
+}
+
+attribute_hidden int Rasprintf_malloc(char **str, const char *fmt, ...)
+{
+    va_list ap;
+    int ret;
+    char dummy[1];
+
+    *str = NULL;
+
+    va_start(ap, fmt);
+    /* could optimize by using non-zero initial size, large
+       enough so that most prints with fill */
+    /* trio does not accept NULL as str */
+    ret = vsnprintf(dummy, 0, fmt, ap); 
+    va_end(ap);
+
+    if (ret <= 0)
+	/* error or empty print */
+	return ret;
+
+    size_t needed = ret + 1;
+    char *buf = (char *) malloc(needed);
+    if (!buf) {
+	errno = ENOMEM;
+	return -1;
+    }
+
+    va_start(ap, fmt);
+    ret = vsnprintf(buf, needed, fmt, ap);
+    va_end(ap);
+
+    if (ret < 0 || (size_t)ret >= needed)
+	/* error */
+	free(buf);
+    else
+	*str = buf;
+    return ret;
+}
+ 
