@@ -1,6 +1,6 @@
 /*
  *  R : A Computer Language for Statistical Data Analysis
- *  Copyright (C) 2001--2013  The R Core Team.
+ *  Copyright (C) 2001--2023  The R Core Team.
  *  Copyright (C) 2003--2010  The R Foundation
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -15,7 +15,7 @@
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, a copy is available at
- *  http://www.r-project.org/Licenses/
+ *  https://www.R-project.org/Licenses/
  */
 
 /* Interface routines, callable from R using .Internal, for Lapack code */
@@ -25,13 +25,32 @@
 #endif
 
 #include <Defn.h>
-#include <ctype.h> /* for toupper */
+#include <ctype.h>  /* for toupper */
+#include <limits.h> /* for PATH_MAX */
+#include <stdlib.h> /* for realpath */
+#include <string.h> /* for strcpy */
+
+#ifdef HAVE_UNISTD_H
+# include <unistd.h> /* for realpath on some systems */
+#endif
+
+#ifdef HAVE_DLFCN_H
+# include <dlfcn.h>  /* for dladdr */
+#endif
+
+#if defined(HAVE_REALPATH) && defined(HAVE_DECL_REALPATH) && !HAVE_DECL_REALPATH
+extern char *realpath(const char *path, char *resolved_path);
+#endif
+
+#if defined(HAVE_DLADDR) && defined(HAVE_DECL_DLADDR) && !HAVE_DECL_DLADDR
+extern int dladdr(void *addr, Dl_info *info);
+#endif
 
 #include "Lapack.h"
 
 /* NB: the handling of dims is odd here.  Most are coerced to be
  * integers (which dimgets currently guarantees), but a couple were
- * used unchecked. 
+ * used unchecked.
  */
 
 /* FIXME: MM would want to make these available to packages;
@@ -78,29 +97,26 @@ static char La_rcond_type(const char *typstr)
 /* La.svd, called from svd */
 static SEXP La_svd(SEXP jobu, SEXP x, SEXP s, SEXP u, SEXP vt)
 {
-    int n, p, info = 0;
-
     if (!isString(jobu))
 	error("'jobu' must be a character string");
-    int *xdims = INTEGER(coerceVector(getAttrib(x, R_DimSymbol), INTSXP));
-    n = xdims[0]; p = xdims[1];
+    int *xdims = INTEGER(coerceVector(getAttrib(x, R_DimSymbol), INTSXP)),
+	n = xdims[0], p = xdims[1], nprot = 2;
 
     /* work on a copy of x  */
     double *xvals;
     if (!isReal(x)) {
-       x = coerceVector(x, REALSXP);
-       xvals = REAL(x);
+	x = PROTECT(coerceVector(x, REALSXP)); nprot++;
+	xvals = REAL(x);
     } else {
 	xvals = (double *) R_alloc(n * (size_t) p, sizeof(double));
 	Memcpy(xvals, REAL(x), n * (size_t) p);
     }
-    PROTECT(x);
 
     SEXP dims = getAttrib(u, R_DimSymbol);
-    if (TYPEOF(dims) != INTSXP) error("non-integer dims");
+    if (TYPEOF(dims) != INTSXP) error("non-integer dim(u)");
     int ldu = INTEGER(dims)[0];
     dims = getAttrib(vt, R_DimSymbol);
-    if (TYPEOF(dims) != INTSXP) error("non-integer dims");
+    if (TYPEOF(dims) != INTSXP) error("non-integer dim(vt)");
     int ldvt = INTEGER(dims)[0];
     double tmp;
     /* min(n,p) large is implausible, but cast to be sure */
@@ -108,17 +124,17 @@ static SEXP La_svd(SEXP jobu, SEXP x, SEXP s, SEXP u, SEXP vt)
 
     /* ask for optimal size of work array */
     const char *ju = CHAR(STRING_ELT(jobu, 0));
-    int lwork = -1;
+    int info = 0, lwork = -1;
     F77_CALL(dgesdd)(ju, &n, &p, xvals, &n, REAL(s),
 		     REAL(u), &ldu, REAL(vt), &ldvt,
-		     &tmp, &lwork, iwork, &info);
+		     &tmp, &lwork, iwork, &info FCONE);
     if (info != 0)
 	error(_("error code %d from Lapack routine '%s'"), info, "dgesdd");
     lwork = (int) tmp;
     double *work = (double *) R_alloc(lwork, sizeof(double));
     F77_CALL(dgesdd)(ju, &n, &p, xvals, &n, REAL(s),
 		     REAL(u), &ldu, REAL(vt), &ldvt,
-		     work, &lwork, iwork, &info);
+		     work, &lwork, iwork, &info FCONE);
     if (info != 0)
 	error(_("error code %d from Lapack routine '%s'"), info, "dgesdd");
 
@@ -131,7 +147,7 @@ static SEXP La_svd(SEXP jobu, SEXP x, SEXP s, SEXP u, SEXP vt)
     SET_VECTOR_ELT(val, 0, s);
     SET_VECTOR_ELT(val, 1, u);
     SET_VECTOR_ELT(val, 2, vt);
-    UNPROTECT(3);
+    UNPROTECT(nprot);
     return val;
 }
 
@@ -177,7 +193,7 @@ static SEXP La_rs(SEXP x, SEXP only_values)
     F77_CALL(dsyevr)(jobv, range, uplo, &n, rx, &n,
 		     &vl, &vu, &il, &iu, &abstol, &m, rvalues,
 		     rz, &n, isuppz,
-		     &tmp, &lwork, &itmp, &liwork, &info);
+		     &tmp, &lwork, &itmp, &liwork, &info FCONE FCONE FCONE);
     if (info != 0)
 	error(_("error code %d from Lapack routine '%s'"), info, "dsyevr");
     lwork = (int) tmp;
@@ -188,7 +204,7 @@ static SEXP La_rs(SEXP x, SEXP only_values)
     F77_CALL(dsyevr)(jobv, range, uplo, &n, rx, &n,
 		     &vl, &vu, &il, &iu, &abstol, &m, rvalues,
 		     rz, &n, isuppz,
-		     work, &lwork, iwork, &liwork, &info);
+		     work, &lwork, iwork, &liwork, &info FCONE FCONE FCONE);
     if (info != 0)
 	error(_("error code %d from Lapack routine '%s'"), info, "dsyevr");
 
@@ -198,7 +214,6 @@ static SEXP La_rs(SEXP x, SEXP only_values)
 	nm = PROTECT(allocVector(STRSXP, 2));
 	SET_STRING_ELT(nm, 1, mkChar("vectors"));
 	SET_VECTOR_ELT(ret, 1, z);
-	UNPROTECT_PTR(z);
     } else {
 	ret = PROTECT(allocVector(VECSXP, 1));
 	nm = PROTECT(allocVector(STRSXP, 1));
@@ -206,7 +221,7 @@ static SEXP La_rs(SEXP x, SEXP only_values)
     SET_STRING_ELT(nm, 0, mkChar("values"));
     setAttrib(ret, R_NamesSymbol, nm);
     SET_VECTOR_ELT(ret, 0, values);
-    UNPROTECT(4);
+    UNPROTECT(ov ? 4 : 5);
     return ret;
 }
 
@@ -270,13 +285,13 @@ static SEXP La_rg(SEXP x, SEXP only_values)
     /* ask for optimal size of work array */
     lwork = -1;
     F77_CALL(dgeev)(jobVL, jobVR, &n, xvals, &n, wR, wI,
-		    left, &n, right, &n, &tmp, &lwork, &info);
+		    left, &n, right, &n, &tmp, &lwork, &info FCONE FCONE);
     if (info != 0)
 	error(_("error code %d from Lapack routine '%s'"), info, "dgeev");
     lwork = (int) tmp;
     work = (double *) R_alloc(lwork, sizeof(double));
     F77_CALL(dgeev)(jobVL, jobVR, &n, xvals, &n, wR, wI,
-		    left, &n, right, &n, work, &lwork, &info);
+		    left, &n, right, &n, work, &lwork, &info FCONE FCONE);
     if (info != 0)
 	error(_("error code %d from Lapack routine '%s'"), info, "dgeev");
 
@@ -337,7 +352,7 @@ static SEXP La_dlange(SEXP A, SEXP type)
 
     SEXP val = PROTECT(allocVector(REALSXP, 1));
     if(*typNorm == 'I') work = (double *) R_alloc(m, sizeof(double));
-    REAL(val)[0] = F77_CALL(dlange)(typNorm, &m, &n, REAL(A), &m, work);
+    REAL(val)[0] = F77_CALL(dlange)(typNorm, &m, &n, REAL(A), &m, work FCONE);
 
     UNPROTECT(nprot);
     return val;
@@ -367,7 +382,7 @@ static SEXP La_dgecon(SEXP A, SEXP norm)
 			      sizeof(double));
     iwork = (int *) R_alloc(m, sizeof(int));
 
-    anorm = F77_CALL(dlange)(typNorm, &m, &n, REAL(A), &m, work);
+    anorm = F77_CALL(dlange)(typNorm, &m, &n, REAL(A), &m, work FCONE);
 
     /* Compute the LU-decomposition and overwrite 'A' with result :*/
     F77_CALL(dgetrf)(&m, &n, REAL(A), &m, iwork, &info);
@@ -389,7 +404,7 @@ static SEXP La_dgecon(SEXP A, SEXP norm)
     }
     F77_CALL(dgecon)(typNorm, &n, REAL(A), &n, &anorm,
 		     /* rcond = */ REAL(val),
-		     work, iwork, &info);
+		     work, iwork, &info FCONE);
     UNPROTECT(2);
     if (info) error(_("error [%d] from Lapack 'dgecon()'"), info);
     return val;
@@ -423,7 +438,7 @@ static SEXP La_dtrcon(SEXP A, SEXP norm)
 		     REAL(val),
 		     /* work : */ (double *) R_alloc(3*(size_t)n, sizeof(double)),
 		     /* iwork: */ (int *)    R_alloc(n, sizeof(int)),
-		     &info);
+		     &info FCONE FCONE FCONE);
     UNPROTECT(nprot);
     if (info) error(_("error [%d] from Lapack 'dtrcon()'"), info);
     return val;
@@ -450,7 +465,7 @@ static SEXP La_zgecon(SEXP A, SEXP norm)
     SEXP val = PROTECT(allocVector(REALSXP, 1));
 
     rwork = (double *) R_alloc(2*(size_t)n, sizeof(Rcomplex));
-    anorm = F77_CALL(zlange)(typNorm, &n, &n, COMPLEX(A), &n, rwork);
+    anorm = F77_CALL(zlange)(typNorm, &n, &n, COMPLEX(A), &n, rwork FCONE);
 
     /* Compute the LU-decomposition and overwrite 'x' with result;
      * working on a copy of A : */
@@ -475,7 +490,7 @@ static SEXP La_zgecon(SEXP A, SEXP norm)
     F77_CALL(zgecon)(typNorm, &n, avals, &n, &anorm,
 		     /* rcond = */ REAL(val),
 		     /* work : */ (Rcomplex *) R_alloc(4*(size_t)n, sizeof(Rcomplex)),
-		     rwork, &info);
+		     rwork, &info FCONE);
     UNPROTECT(1);
     if (info) error(_("error [%d] from Lapack 'zgecon()'"), info);
     return val;
@@ -511,7 +526,7 @@ static SEXP La_ztrcon(SEXP A, SEXP norm)
 		     REAL(val),
 		     /* work : */ (Rcomplex *) R_alloc(2*(size_t)n, sizeof(Rcomplex)),
 		     /* rwork: */ (double *)   R_alloc(n, sizeof(double)),
-		     &info);
+		     &info FCONE FCONE FCONE);
     UNPROTECT(1);
     if (info) error(_("error [%d] from Lapack 'ztrcon()'"), info);
     return val;
@@ -522,7 +537,7 @@ static SEXP La_ztrcon(SEXP A, SEXP norm)
 }
 
 /* Complex case of solve.default: see the comments in La_solve */
-static SEXP La_solve_cmplx(SEXP A, SEXP Bin)
+static SEXP La_solve_cmplx(SEXP A, SEXP Bin, SEXP tolin)
 {
 #ifdef HAVE_FORTRAN_DOUBLE_COMPLEX
     int n, p, info, *ipiv, *Adims, *Bdims;
@@ -533,6 +548,7 @@ static SEXP La_solve_cmplx(SEXP A, SEXP Bin)
     Adims = INTEGER(coerceVector(getAttrib(A, R_DimSymbol), INTSXP));
     n = Adims[0];
     if(n == 0) error(_("'a' is 0-diml"));
+    size_t nl = n;
     int n2 = Adims[1];
     if(n2 != n) error(_("'a' (%d x %d) must be square"), n, n2);
     Adn = getAttrib(A, R_DimNamesSymbol);
@@ -558,12 +574,12 @@ static SEXP La_solve_cmplx(SEXP A, SEXP Bin)
 	p = 1;
 	if(length(Bin) != n)
 	    error(_("'b' (%d x %d) must be compatible with 'a' (%d x %d)"),
-		  length(Bin), p, n, n);	
+		  length(Bin), p, n, n);
 	PROTECT(B = allocVector(CPLXSXP, n));
 	if (!isNull(Adn)) setAttrib(B, R_NamesSymbol, VECTOR_ELT(Adn, 1));
     }
     Bin = PROTECT(coerceVector(Bin, CPLXSXP));
-    Memcpy(COMPLEX(B), COMPLEX(Bin), (size_t)n * p);
+    Memcpy(COMPLEX(B), COMPLEX(Bin), nl * p);
 
     ipiv = (int *) R_alloc(n, sizeof(int));
 
@@ -572,8 +588,8 @@ static SEXP La_solve_cmplx(SEXP A, SEXP Bin)
 	A = coerceVector(A, CPLXSXP);
 	avals = COMPLEX(A);
     } else {
-	avals = (Rcomplex *) R_alloc((size_t)n * n, sizeof(Rcomplex));
-	Memcpy(avals, COMPLEX(A), (size_t) n * n);
+	avals = (Rcomplex *) R_alloc(nl * nl, sizeof(Rcomplex));
+	Memcpy(avals, COMPLEX(A), nl * nl);
     }
     PROTECT(A);
     F77_CALL(zgesv)(&n, &p, avals, &n, ipiv, COMPLEX(B), &n, &info);
@@ -581,7 +597,24 @@ static SEXP La_solve_cmplx(SEXP A, SEXP Bin)
 	error(_("argument %d of Lapack routine %s had invalid value"),
 	      -info, "zgesv");
     if (info > 0)
-	error(("Lapack routine zgesv: system is exactly singular"));
+	error(_("Lapack routine zgesv: system is exactly singular"));
+    int OK = 1;
+    for (size_t i = 0; i < nl*nl; i++)
+	if (!isfinite(avals[i].r) || !isfinite(avals[i].i)) {OK = 0; break;}
+    double tol = asReal(tolin);
+    if(OK == 1 && tol > 0) {
+	char one[2] = "1";
+	double anorm = F77_CALL(zlange)(one, &n, &n, COMPLEX(A), &n,
+					(double*) NULL FCONE);
+	Rcomplex *work = (Rcomplex *) R_alloc(2*nl, sizeof(Rcomplex));
+	double *rwork = (double *) R_alloc(2*nl, sizeof(double));
+	double rcond;
+	F77_CALL(zgecon)(one, &n, avals, &n, &anorm, &rcond, work, rwork,
+			 &info FCONE);
+	if (rcond < tol)
+	    error(_("system is computationally singular: reciprocal condition number = %g"),
+		  rcond);
+    }
     UNPROTECT(3);  /* B, Bin, A */
     return B;
 #else
@@ -672,18 +705,19 @@ static SEXP qr_coef_cmplx(SEXP Q, SEXP Bin)
     lwork = -1;
     F77_CALL(zunmqr)("L", "C", &n, &nrhs, &k,
 		     COMPLEX(qr), &n, COMPLEX(tau), COMPLEX(B), &n,
-		     &tmp, &lwork, &info);
+		     &tmp, &lwork, &info FCONE FCONE);
     if (info != 0)
 	error(_("error code %d from Lapack routine '%s'"), info, "zunmqr");
     lwork = (int) tmp.r;
     work = (Rcomplex *) R_alloc(lwork, sizeof(Rcomplex));
     F77_CALL(zunmqr)("L", "C", &n, &nrhs, &k,
 		     COMPLEX(qr), &n, COMPLEX(tau), COMPLEX(B), &n,
-		     work, &lwork, &info);
+		     work, &lwork, &info FCONE FCONE);
     if (info != 0)
 	error(_("error code %d from Lapack routine '%s'"), info, "zunmqr");
     F77_CALL(ztrtrs)("U", "N", "N", &k, &nrhs,
-		     COMPLEX(qr), &n, COMPLEX(B), &n, &info);
+		     COMPLEX(qr), &n, COMPLEX(B), &n, &info
+		     FCONE FCONE FCONE);
     if (info != 0)
 	error(_("error code %d from Lapack routine '%s'"), info, "ztrtrs");
     UNPROTECT(1);
@@ -718,14 +752,14 @@ static SEXP qr_qy_cmplx(SEXP Q, SEXP Bin, SEXP trans)
     lwork = -1;
     F77_CALL(zunmqr)("L", tr ? "C" : "N", &n, &nrhs, &k,
 		     COMPLEX(qr), &n, COMPLEX(tau), COMPLEX(B), &n,
-		     &tmp, &lwork, &info);
+		     &tmp, &lwork, &info FCONE FCONE);
     if (info != 0)
 	error(_("error code %d from Lapack routine '%s'"), info, "zunmqr");
     lwork = (int) tmp.r;
     work = (Rcomplex *) R_alloc(lwork, sizeof(Rcomplex));
     F77_CALL(zunmqr)("L", tr ? "C" : "N", &n, &nrhs, &k,
 		     COMPLEX(qr), &n, COMPLEX(tau), COMPLEX(B), &n,
-		     work, &lwork, &info);
+		     work, &lwork, &info FCONE FCONE);
     if (info != 0)
 	error(_("error code %d from Lapack routine '%s'"), info, "zunmqr");
     UNPROTECT(1);
@@ -774,14 +808,14 @@ static SEXP La_svd_cmplx(SEXP jobu, SEXP x, SEXP s, SEXP u, SEXP v)
     ldv = INTEGER(dims)[0];
     F77_CALL(zgesdd)(jz, &n, &p, xvals, &n, REAL(s),
 		     COMPLEX(u), &ldu, COMPLEX(v), &ldv,
-		     &tmp, &lwork, rwork, iwork, &info);
+		     &tmp, &lwork, rwork, iwork, &info FCONE);
     if (info != 0)
 	error(_("error code %d from Lapack routine '%s'"), info, "zgesdd");
     lwork = (int) tmp.r;
     Rcomplex *work = (Rcomplex *) R_alloc(lwork, sizeof(Rcomplex));
     F77_CALL(zgesdd)(jz, &n, &p, xvals, &n, REAL(s),
 		     COMPLEX(u), &ldu, COMPLEX(v), &ldv,
-		     work, &lwork, rwork, iwork, &info);
+		     work, &lwork, rwork, iwork, &info FCONE);
     if (info != 0)
 	error(_("error code %d from Lapack routine '%s'"), info, "zgesdd");
 
@@ -824,18 +858,18 @@ static SEXP La_rs_cmplx(SEXP xin, SEXP only_values)
     SEXP values = PROTECT(allocVector(REALSXP, n));
     rvalues = REAL(values);
 
-    rwork = (double *) R_alloc((3*(size_t)n-2) > 1 ? 3*(size_t)n-2 : 1, 
+    rwork = (double *) R_alloc((3*(size_t)n-2) > 1 ? 3*(size_t)n-2 : 1,
 			       sizeof(double));
     /* ask for optimal size of work array */
     lwork = -1;
     F77_CALL(zheev)(jobv, uplo, &n, rx, &n, rvalues, &tmp, &lwork, rwork,
-		    &info);
+		    &info FCONE FCONE);
     if (info != 0)
 	error(_("error code %d from Lapack routine '%s'"), info, "zheev");
     lwork = (int) tmp.r;
     work = (Rcomplex *) R_alloc(lwork, sizeof(Rcomplex));
     F77_CALL(zheev)(jobv, uplo, &n, rx, &n, rvalues, work, &lwork, rwork,
-		    &info);
+		    &info FCONE FCONE);
     if (info != 0)
 	error(_("error code %d from Lapack routine '%s'"), info, "zheev");
     SEXP ret, nm;
@@ -889,13 +923,15 @@ static SEXP La_rg_cmplx(SEXP x, SEXP only_values)
     /* ask for optimal size of work array */
     lwork = -1;
     F77_CALL(zgeev)(jobVL, jobVR, &n, xvals, &n, COMPLEX(values),
-		    left, &n, right, &n, &tmp, &lwork, rwork, &info);
+		    left, &n, right, &n, &tmp, &lwork, rwork, &info
+		    FCONE FCONE);
     if (info != 0)
 	error(_("error code %d from Lapack routine '%s'"), info, "zgeev");
     lwork = (int) tmp.r;
     work = (Rcomplex *) R_alloc(lwork, sizeof(Rcomplex));
     F77_CALL(zgeev)(jobVL, jobVR, &n, xvals, &n, COMPLEX(values),
-		    left, &n, right, &n, work, &lwork, rwork, &info);
+		    left, &n, right, &n, work, &lwork, rwork, &info
+		    FCONE FCONE);
     if (info != 0)
 	error(_("error code %d from Lapack routine '%s'"), info, "zgeev");
 
@@ -940,10 +976,10 @@ static SEXP La_chol(SEXP A, SEXP pivot, SEXP stol)
     if (piv != 0 && piv != 1) error("invalid '%s' value", "pivot");
     if(!piv) {
 	int info;
-	F77_CALL(dpotrf)("Upper", &m, REAL(ans), &m, &info);
+	F77_CALL(dpotrf)("U", &m, REAL(ans), &m, &info FCONE);
 	if (info != 0) {
 	    if (info > 0)
-		error(_("the leading minor of order %d is not positive definite"),
+		error(_("the leading minor of order %d is not positive"),
 		      info);
 	    error(_("argument %d of Lapack routine %s had invalid value"),
 		  -info, "dpotrf");
@@ -954,22 +990,24 @@ static SEXP La_chol(SEXP A, SEXP pivot, SEXP stol)
 	int *ip = INTEGER(piv);
 	double *work = (double *) R_alloc(2 * (size_t)m, sizeof(double));
 	int rank, info;
-	F77_CALL(dpstrf)("U", &m, REAL(ans), &m, ip, &rank, &tol, work, &info);
+	F77_CALL(dpstrf)("U", &m, REAL(ans), &m, ip, &rank, &tol, work, &info
+			 FCONE);
 	if (info != 0) {
 	    if (info > 0)
-		warning(_("the matrix is either rank-deficient or indefinite"));
+		warning(_("the matrix is either rank-deficient or not positive definite"));
 	    else
 		error(_("argument %d of Lapack routine %s had invalid value"),
 		      -info, "dpstrf");
 	}
 	setAttrib(ans, install("pivot"), piv);
-	setAttrib(ans, install("rank"), ScalarInteger(rank));
+	SEXP s_rank = install("rank");
+	setAttrib(ans, s_rank, ScalarInteger(rank));
 	SEXP cn, dn = getAttrib(ans, R_DimNamesSymbol);
 	if (!isNull(dn) && !isNull(cn = VECTOR_ELT(dn, 1))) {
 	    // need to pivot the colnames
 	    SEXP dn2 = PROTECT(duplicate(dn));
 	    SEXP cn2 = VECTOR_ELT(dn2, 1);
-	    for(int i = 0; i < m; i++) 
+	    for(int i = 0; i < m; i++)
 		SET_STRING_ELT(cn2, i, STRING_ELT(cn, ip[i] - 1)); // base 1
 	    setAttrib(ans, R_DimNamesSymbol, dn2);
 	    UNPROTECT(1);
@@ -1008,7 +1046,7 @@ static SEXP La_chol2inv(SEXP A, SEXP size)
 		REAL(ans)[i + j * SZ] = REAL(Amat)[i + j * M];
 	}
 	int info;
-	F77_CALL(dpotri)("Upper", &sz, REAL(ans), &sz, &info);
+	F77_CALL(dpotri)("U", &sz, REAL(ans), &sz, &info FCONE);
 	if (info != 0) {
 	    UNPROTECT(nprot);
 	    if (info > 0)
@@ -1031,15 +1069,16 @@ static SEXP La_chol2inv(SEXP A, SEXP size)
 static SEXP La_solve(SEXP A, SEXP Bin, SEXP tolin)
 {
     int n, p;
-    double *avals, anorm, rcond, tol = asReal(tolin), *work;
+    double *avals,  tol = asReal(tolin);
     SEXP B, Adn, Bdn;
 
-    if (!(isMatrix(A) && 
+    if (!(isMatrix(A) &&
 	  (TYPEOF(A) == REALSXP || TYPEOF(A) == INTSXP || TYPEOF(A) == LGLSXP)))
 	error(_("'a' must be a numeric matrix"));
     int *Adims = INTEGER(coerceVector(getAttrib(A, R_DimSymbol), INTSXP));
     n = Adims[0];
     if(n == 0) error(_("'a' is 0-diml"));
+    size_t nl = n;
     int n2 = Adims[1];
     if(n2 != n) error(_("'a' (%d x %d) must be square"), n, n2);
     Adn = getAttrib(A, R_DimNamesSymbol);
@@ -1067,13 +1106,13 @@ static SEXP La_solve(SEXP A, SEXP Bin, SEXP tolin)
 	p = 1;
 	if(length(Bin) != n)
 	    error(_("'b' (%d x %d) must be compatible with 'a' (%d x %d)"),
-		  length(Bin), p, n, n);	
+		  length(Bin), p, n, n);
 	PROTECT(B = allocVector(REALSXP, n));
 	if (!isNull(Adn)) setAttrib(B, R_NamesSymbol, VECTOR_ELT(Adn, 1));
     }
     PROTECT(Bin = coerceVector(Bin, REALSXP));
-    Memcpy(REAL(B), REAL(Bin), (size_t)n * p);
-    
+    Memcpy(REAL(B), REAL(Bin), nl * p);
+
     int *ipiv = (int *) R_alloc(n, sizeof(int));
 
     /* work on a copy of A */
@@ -1081,8 +1120,8 @@ static SEXP La_solve(SEXP A, SEXP Bin, SEXP tolin)
 	A = coerceVector(A, REALSXP);
 	avals = REAL(A);
     } else {
-	avals = (double *) R_alloc((size_t)n * n, sizeof(double));
-	Memcpy(avals, REAL(A), (size_t)n * n);
+	avals = (double *) R_alloc(nl * nl, sizeof(double));
+	Memcpy(avals, REAL(A), nl * nl);
     }
     PROTECT(A);
     int info;
@@ -1093,11 +1132,18 @@ static SEXP La_solve(SEXP A, SEXP Bin, SEXP tolin)
     if (info > 0)
 	error(_("Lapack routine %s: system is exactly singular: U[%d,%d] = 0"),
 	      "dgesv", info, info);
-    if(tol > 0) {
+    // LAPACK 3.11.0 fails here if A contains NaNs
+    int OK = 1;
+    for (size_t i = 0; i < nl*nl; i++)
+	if (!isfinite(avals[i])) {OK = 0; break;}
+    if(OK == 1 && tol > 0) {
 	char one[2] = "1";
-	anorm = F77_CALL(dlange)(one, &n, &n, REAL(A), &n, (double*) NULL);
-	work = (double *) R_alloc(4*(size_t)n, sizeof(double));
-	F77_CALL(dgecon)(one, &n, avals, &n, &anorm, &rcond, work, ipiv, &info);
+	double rcond;
+	double anorm = F77_CALL(dlange)(one, &n, &n, REAL(A), &n,
+					(double*) NULL FCONE);
+	double *work = (double *) R_alloc(4*nl, sizeof(double));
+	F77_CALL(dgecon)(one, &n, avals, &n, &anorm, &rcond, work, ipiv,
+			 &info FCONE);
 	if (rcond < tol)
 	    error(_("system is computationally singular: reciprocal condition number = %g"),
 		  rcond);
@@ -1125,7 +1171,7 @@ static SEXP La_qr(SEXP Ain)
 
     SEXP jpvt = PROTECT(allocVector(INTSXP, n));
     for (int i = 0; i < n; i++) INTEGER(jpvt)[i] = 0;
-    SEXP tau = PROTECT(allocVector(REALSXP, m < n ? m : n));
+    SEXP tau = PROTECT(allocVector(REALSXP, m < n ? m : n)); // qraux
     int info, lwork = -1;
     double tmp;
     F77_CALL(dgeqp3)(&m, &n, REAL(A), &m, INTEGER(jpvt), REAL(tau),
@@ -1166,35 +1212,34 @@ static SEXP La_qr(SEXP Ain)
 /* Real case of qr.coef */
 static SEXP qr_coef_real(SEXP Q, SEXP Bin)
 {
-    int n, nrhs, k, *Bdims;
-    SEXP B, qr = VECTOR_ELT(Q, 0), tau = VECTOR_ELT(Q, 2);
-    double *work, tmp;
-
-    k = LENGTH(tau);
     if (!isMatrix(Bin)) error(_("'b' must be a numeric matrix"));
-    
-    B = PROTECT(isReal(Bin) ? duplicate(Bin) : coerceVector(Bin, REALSXP));
 
-    n = INTEGER(coerceVector(getAttrib(qr, R_DimSymbol), INTSXP))[0];
-    Bdims = INTEGER(coerceVector(getAttrib(B, R_DimSymbol), INTSXP));
+    SEXP B = PROTECT(isReal(Bin) ? duplicate(Bin) : coerceVector(Bin, REALSXP)),
+	qr  = VECTOR_ELT(Q, 0), // qr$qr
+	tau = VECTOR_ELT(Q, 2); // qr$qraux
+    int k = LENGTH(tau),
+	n =      INTEGER(coerceVector(getAttrib(qr, R_DimSymbol), INTSXP))[0],
+	*Bdims = INTEGER(coerceVector(getAttrib(B,  R_DimSymbol), INTSXP));
     if(Bdims[0] != n)
 	error(_("right-hand side should have %d not %d rows"), n, Bdims[0]);
-    nrhs = Bdims[1];
-    int lwork = -1, info;
+    int nrhs = Bdims[1],
+	lwork = -1, info;
+    double tmp;
     F77_CALL(dormqr)("L", "T", &n, &nrhs, &k,
 		     REAL(qr), &n, REAL(tau), REAL(B), &n,
-		     &tmp, &lwork, &info);
+		     &tmp, &lwork, &info FCONE FCONE);
     if (info != 0)
-	error(_("error code %d from Lapack routine '%s'"), info, "dormqr");
+	error(_("error code %d from Lapack routine '%s'"), info, "dormqr [tmp]");
     lwork = (int) tmp;
-    work = (double *) R_alloc(lwork, sizeof(double));
+    double *work = (double *) R_alloc(lwork, sizeof(double));
     F77_CALL(dormqr)("L", "T", &n, &nrhs, &k,
 		     REAL(qr), &n, REAL(tau), REAL(B), &n,
-		     work, &lwork, &info);
+		     work, &lwork, &info FCONE FCONE);
     if (info != 0)
-	error(_("error code %d from Lapack routine '%s'"), info, "dormqr");
+	error(_("error code %d from Lapack routine '%s'"), info, "dormqr [work]");
     F77_CALL(dtrtrs)("U", "N", "N", &k, &nrhs,
-		     REAL(qr), &n, REAL(B), &n, &info);
+		     REAL(qr), &n, REAL(B), &n, &info
+		     FCONE FCONE FCONE);
     if (info != 0)
 	error(_("error code %d from Lapack routine '%s'"), info, "dtrtrs");
     UNPROTECT(1);
@@ -1222,14 +1267,14 @@ static SEXP qr_qy_real(SEXP Q, SEXP Bin, SEXP trans)
     int lwork = -1, info;
     F77_CALL(dormqr)("L", tr ? "T" : "N", &n, &nrhs, &k,
 		     REAL(qr), &n, REAL(tau), REAL(B), &n,
-		     &tmp, &lwork, &info);
+		     &tmp, &lwork, &info FCONE FCONE);
     if (info != 0)
 	error(_("error code %d from Lapack routine '%s'"), info, "dormqr");
     lwork = (int) tmp;
     work = (double *) R_alloc(lwork, sizeof(double));
     F77_CALL(dormqr)("L", tr ? "T" : "N", &n, &nrhs, &k,
 		     REAL(qr), &n, REAL(tau), REAL(B), &n,
-		     work, &lwork, &info);
+		     work, &lwork, &info FCONE FCONE);
     if (info != 0)
 	error(_("error code %d from Lapack routine '%s'"), info, "dormqr");
     UNPROTECT(1);
@@ -1287,7 +1332,8 @@ static SEXP det_ge_real(SEXP Ain, SEXP logarithm)
     SET_STRING_ELT(nm, 1, mkChar("sign"));
     setAttrib(val, R_NamesSymbol, nm);
     SET_VECTOR_ELT(val, 0, ScalarReal(modulus));
-    setAttrib(VECTOR_ELT(val, 0), install("logarithm"), ScalarLogical(useLog));
+    SEXP s_logarithm = install("logarithm");
+    setAttrib(VECTOR_ELT(val, 0), s_logarithm, ScalarLogical(useLog));
     SET_VECTOR_ELT(val, 1, ScalarInteger(sign));
     setAttrib(val, R_ClassSymbol, ScalarString(mkChar("det")));
     UNPROTECT(3);
@@ -1311,7 +1357,7 @@ static SEXP mod_do_lapack(SEXP call, SEXP op, SEXP args, SEXP env)
     case 8: ans = La_dtrcon(CAR(args), CADR(args)); break;
     case 9: ans = La_zgecon(CAR(args), CADR(args)); break;
     case 10: ans = La_ztrcon(CAR(args), CADR(args)); break;
-    case 11: ans = La_solve_cmplx(CAR(args), CADR(args)); break;
+    case 11: ans = La_solve_cmplx(CAR(args), CADR(args), CADDR(args)); break;
 
     case 100: ans = La_solve(CAR(args), CADR(args), CADDR(args)); break;
     case 101: ans = La_qr(CAR(args)); break;
@@ -1345,7 +1391,7 @@ static SEXP mod_do_lapack(SEXP call, SEXP op, SEXP args, SEXP env)
 	ans = La_svd_cmplx(a1, a2, a3, a4, CAR(args));
 	break;
     }
-    case 1000:
+    case 1000: // La_version()
     {
 	int major, minor, patch;
 	char str[20];
@@ -1354,8 +1400,34 @@ static SEXP mod_do_lapack(SEXP call, SEXP op, SEXP args, SEXP env)
 	ans = mkString(str);
 	break;
     }
+    case 1001: // La_library()
+    {
+#if defined(HAVE_DLADDR) && defined(HAVE_REALPATH)
+	Dl_info dl_info;
+	/* the call to dladdr() converts a function pointer to an object
+	   pointer, which is not allowed by ISO C, but there is no compliant
+	   alternative to using dladdr() */
+	// dladdr has first arg void * on Solaris.  This is not POSIX.
+	if (dladdr((void *) F77_NAME(ilaver), &dl_info)) {
+	    char buf[PATH_MAX+1];
+	    char *res = realpath(dl_info.dli_fname, buf);
+	    if (res) {
+		SEXP nfo = R_NilValue;
+		if (strstr(res, "flexiblas"))
+		    nfo = R_flexiblas_info();
+		if (isNull(nfo))
+		    nfo = mkChar(res);
+		ans = ScalarString(nfo);
+		break;
+	    }
+	}
+#endif
+	ans = mkString(""); /* LAPACK library not known */
+	break;
     }
 
+    }// switch
+
     return ans;
 }
 
@@ -1369,7 +1441,7 @@ void
 R_init_lapack(DllInfo *info)
 {
     R_LapackRoutines *tmp;
-    tmp = Calloc(1, R_LapackRoutines);
+    tmp = R_Calloc(1, R_LapackRoutines);
 
     tmp->do_lapack = mod_do_lapack;
     R_setLapackRoutines(tmp);
